from mlkem.math.constants import n, q
from mlkem.math.field import Zm

BITS_IN_BYTE = 8
MAX_D = q.bit_length()


def bits_to_bytes(bits: list[int]) -> list[int]:
    """Converts a bit array (of a length that is a multiple of 8) into an array of bytes.

    Bytes are represented as unsigned numbers in the range [0, 255]. Bits are either 0 or 1.

    Args:
        | bits (:type:`list[int]`): The bit array (of a length that is a multiple of 8).

    Returns:
        :type:`list[int]`: The array of bytes equivalent to the bit array.
    """
    length = len(bits)
    if length % BITS_IN_BYTE != 0:
        raise ValueError(
            f"Bit array must have a length that is a multiple of 8 (got {length})."
        )

    result = [0 for _ in range(length // BITS_IN_BYTE)]
    for i in range(length):
        bitval = bits[i] * (1 << (i % BITS_IN_BYTE))
        result[i // BITS_IN_BYTE] = result[i // BITS_IN_BYTE] + bitval

    return result


def bytes_to_bits(byts: list[int]) -> list[int]:
    """Converts a byte array into an array of bits.

    Bytes are represented as unsigned numbers in the range [0, 255]. Bits are either 0 or 1.

    Args:
        | byts (:type:`list[int]`): The byte array.

    Returns:
        :type:`list[int]`: The array of bits equivalent to the byte array.
    """
    c = byts.copy()

    result: list[int] = []
    for i in range(len(c)):
        for _ in range(BITS_IN_BYTE):
            result.append(c[i] & 1)
            c[i] //= 2

    return result


def _round_fraction(x: int, y: int) -> int:
    """Round the fraction x/y to the nearest integer."""
    return (2 * x + y) // (2 * y)


def compress(d: int, x: Zm) -> Zm:
    if not d < MAX_D:
        raise ValueError(f"d must be less than {MAX_D} (got {d}).")
    if x.m != q:
        raise ValueError(f"Element being compressed must be in Z_q (got Z_{x.m}).")

    m = 1 << d
    val = _round_fraction(m * x.val, q) % m
    return Zm(val, m)


def decompress(d: int, y: Zm) -> Zm:
    if not d < MAX_D:
        raise ValueError(f"d must be less than {MAX_D} (got {d}).")

    m = 1 << d
    val = _round_fraction(q * y.val, m)
    return Zm(val, q)


def byte_encode(d: int, f: list[Zm]) -> bytes:
    if len(f) != n:
        raise ValueError(f"f must have {n} elements (got {len(f)}).")

    if d > MAX_D or d < 1:
        raise ValueError(f"d may not be greater than {MAX_D} or less than 1 (got {d}).")

    b = [0 for _ in range(n * d)]
    for i in range(n):
        a = f[i].val

        for j in range(d):
            x = a & 1
            b[i * d + j] = x
            a = (a - x) // 2

    return bytes(bits_to_bytes(b))


def byte_decode(d: int, b: bytes) -> list[Zm]:
    if d > MAX_D or d < 1:
        raise ValueError(f"d may not be greater than {MAX_D} or less than 1 (got {d}).")

    m = q if d == MAX_D else 1 << d
    bits = bytes_to_bits([x for x in b])

    f = []
    for i in range(n):
        fi = sum([bits[i * d + j] * (1 << j) for j in range(d)])
        fi_m = Zm(fi, m)
        f.append(fi_m)

    return f



from abc import abstractmethod
from typing import Protocol, Self


# the entries of a matrix must have a type supporting addition
# define that interface via this protocol
class Field(Protocol):
    @abstractmethod
    def __add__(self, other: Self) -> Self:
        pass

    @abstractmethod
    def __sub__(self, other: Self) -> Self:
        pass

    @abstractmethod
    def __mul__(self, other: Self) -> Self:
        pass



from os import urandom

from mlkem.auxiliary.crypto import g, h, j
from mlkem.auxiliary.general import byte_decode, byte_encode
from mlkem.fast_k_pke import Fast_K_PKE
from mlkem.fastmath import byte_decode_matrix, byte_encode_matrix  # type: ignore
from mlkem.k_pke import K_PKE
from mlkem.parameter_set import ParameterSet


class ML_KEM:
    def __init__(self, parameters: ParameterSet, fast: bool = True):
        self.parameters = parameters
        self.fast = fast
        self.k_pke = Fast_K_PKE(parameters) if fast else K_PKE(parameters)

    def key_gen(self) -> tuple[bytes, bytes]:
        d = urandom(32)
        z = urandom(32)

        return self._key_gen(d, z)

    def encaps(self, ek: bytes) -> tuple[bytes, bytes]:
        self._check_ek(ek)
        m = urandom(32)
        return self._encaps(ek, m)

    def decaps(self, dk: bytes, c: bytes) -> bytes:
        return self._decaps(dk, c)

    def _key_gen(self, d: bytes, z: bytes) -> tuple[bytes, bytes]:
        ek, dk_pke = self.k_pke.key_gen(d)
        dk = dk_pke + ek + h(ek) + z
        return ek, dk

    def _encaps(self, ek: bytes, m: bytes) -> tuple[bytes, bytes]:
        k, r = g(m + h(ek))
        c = self.k_pke.encrypt(ek, m, r)
        return k, c

    def _decaps(self, dk: bytes, c: bytes) -> bytes:
        k = self.parameters.k
        # extract encryption and decryption keys, hash of encryption key, and rejection value
        dk_pke = dk[: 384 * k]
        ek_pke = dk[384 * k : 768 * k + 32]
        h_ = dk[768 * k + 32 : 768 * k + 64]
        z = dk[768 * k + 64 : 768 * k + 96]

        # decrypt ciphertext
        m_prime = self.k_pke.decrypt(dk_pke, c)
        k_prime, r_prime = g(m_prime + h_)
        k_bar = j(z + c)

        # re-encrypt using the derived randomness r_prime
        c_prime = self.k_pke.encrypt(ek_pke, m_prime, r_prime)
        if c != c_prime:
            # if ciphertexts do not match, then implicitly reject
            k_prime = k_bar

        return k_prime

    def _check_ek(self, ek: bytes) -> None:
        k = self.parameters.k

        if len(ek) != 384 * k + 32:
            raise ValueError(f"Expected key of size {384 * k + 32}, got {len(ek)}.")

        if self.fast:
            expected = ek[: 384 * k]
            test = byte_encode_matrix(byte_decode_matrix(ek, 12, k), 12)
            if expected != test:
                raise ValueError(
                    "Encapsulation key contains bytes greater than or equal to q."
                )
        else:
            for i in range(k):
                expected = ek[i * 384 : i * 384 + 384]
                test = byte_encode(12, byte_decode(12, expected))
                if expected != test:
                    raise ValueError(
                        "Encapsulation key contains bytes greater than or equal to q."
                    )



from __future__ import annotations


class Zm:
    r"""Represents elements of the field :math:`\mathbb{Z}_m`."""

    val: int

    def __init__(self, val: int, m: int):
        self.m = m
        self.val = val % m

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Zm):
            return NotImplemented

        return self.m == other.m and self.val == other.val

    def __repr__(self) -> str:
        return f"{self.val}"

    def __add__(self, y: Zm) -> Zm:
        r"""Add two elements x and y where :math:`x, y \in \mathbb{Z}_m`.

        The resulting element will also be in :math:`\mathbb{Z}_m`.

        Args:
            | y (:class:`Zm`): The element to add to x (represented by self).

        Returns:
            :class:`Zm`: An element equal to :math:`x + y \pmod{m}`.
        """
        if self.m != y.m:
            raise ValueError(
                f"Cannot add elements in different fields ({self.m} and {y.m})."
            )
        return Zm(self.val + y.val, self.m)

    def __sub__(self, y: Zm) -> Zm:
        r"""Subtract two elements x and y where :math:`x, y \in \mathbb{Z}_m`.

        The resulting element will also be in :math:`\mathbb{Z}_m`.

        Args:
            | y (:class:`Zm`): The element to subtract from x (represented by self).

        Returns:
            :class:`Zm`: An element equal to :math:`x - y \pmod{m}`.
        """
        return Zm(self.val - y.val, self.m)

    def __mul__(self, y: Zm) -> Zm:
        r"""Multiply two elements x and y where :math:`x, y \in \mathbb{Z}_m`.

        The resulting element will also be in :math:`\mathbb{Z}_m`.

        Args:
            | y (:class:`Zm`): The element to multiply x by (represented by self).

        Returns:
            :class:`Zm`: An element equal to :math:`x * y \pmod{m}`.
        """
        if self.m != y.m:
            raise ValueError(
                f"Cannot add elements in different fields ({self.m} and {y.m})."
            )
        return Zm(self.val * y.val, self.m)



from hashlib import sha3_256, sha3_512, shake_128, shake_256


def prf(eta: int, s: bytes, b: bytes) -> bytes:
    if eta not in {2, 3}:
        raise ValueError(f"eta must be 2 or 3 (got {eta})")
    if len(s) != 32:
        raise ValueError(f"len(s) must be 32 (got {len(s)})")
    if len(b) != 1:
        raise ValueError(f"len(b) must be 1 (got {len(b)})")

    # length passed to digest is byte length, so omit factor of 8 from spec (which uses bit length)
    return shake_256(s + b).digest(64 * eta)


def h(s: bytes) -> bytes:
    return sha3_256(s).digest()


def j(s: bytes) -> bytes:
    # length passed to digest is byte length, so omit factor of 8 from spec (which uses bit length)
    return shake_256(s).digest(32)


def g(c: bytes) -> tuple[bytes, bytes]:
    ab = sha3_512(c).digest()
    return ab[:32], ab[32:]


class XOF:
    def __init__(self) -> None:
        # https://cryptojedi.org/papers/terminate-20230516.pdf
        self.chunk_size = 840
        self.shake = shake_128()
        self.data = b""
        self.idx = 0

    def absorb(self, string: bytes) -> None:
        self.shake.update(string)
        self.data += self.shake.digest(self.chunk_size)

    def squeeze(self, length: int) -> bytes:
        while self.idx + length > len(self.data):
            self.data += self.shake.digest(self.chunk_size)

        result = self.data[self.idx : self.idx + length]
        self.idx += length
        return result



from __future__ import annotations
from typing import Callable, Generic, TypeVar

from mlkem.data_types import Field

T = TypeVar("T", bound=Field)


class Matrix(Generic[T]):
    rows: int
    cols: int
    entries: list[T]

    def __init__(
        self,
        rows: int,
        cols: int,
        entries: list[T] | None = None,
        constructor: Callable[[], T] | None = None,
    ) -> None:
        self.rows = rows
        self.cols = cols

        if entries is not None:
            assert (
                len(entries) == rows * cols
            ), f"Entries had {len(entries)} entries, expected {rows} * {cols} entries."
            self.entries = entries
        elif constructor is not None:
            self.entries = [constructor() for _ in range(rows * cols)]
        else:
            raise ValueError("Must provide either entries or constructor")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Matrix):
            return NotImplemented

        if self.rows != other.rows or self.cols != other.cols:
            return False

        return all([x == y for (x, y) in zip(self.entries, other.entries)])

    def __repr__(self) -> str:
        result = "[ "
        result += ", ".join(
            [
                "[ " + ", ".join([repr(self[(i, j)]) for j in range(self.cols)]) + " ]"
                for i in range(self.rows)
            ]
        )
        result += " ]"
        return result

    def __getitem__(self, index: tuple[int, int]) -> T:
        r"""Get the element at index (i, j) of the matrix.

        The elements of the matrix are interpreted in row-major order. Since the matrix
        is represented as a flat list, this means that each row advances the index by the
        amount of columns in the matrix.

        Args:
            | index (:type:`tuple[int, int]`): A tuple of (row index, column index).

        Returns:
            :type:`T`: The element at index (i, j) of the matrix.
        """
        row, col = index
        if row >= self.rows or col >= self.cols:
            raise IndexError(
                f"{row}, {col} is not a valid index for a {self.rows}x{self.cols} matrix."
            )

        i = row * self.cols + col
        return self.entries[i]

    def __setitem__(self, index: tuple[int, int], value: T) -> None:
        r"""Set the element at index (i, j) of the matrix.

        The elements of the matrix are interpreted in row-major order. Since the matrix
        is represented as a flat list, this means that each row advances the index by the
        amount of columns in the matrix.

        Args:
            | index (:type:`tuple[int, int]`): A tuple of (row index, column index).
            | value (:type:`T`): The value to set at the given index.
        """
        row, col = index
        if row >= self.rows or col >= self.cols:
            raise IndexError(
                f"{row}, {col} is not a valid index for a {self.rows}x{self.cols} matrix."
            )

        i = row * self.cols + col
        self.entries[i] = value

    def __add__(self, other: Matrix[T]) -> Matrix[T]:
        r"""Add two matrices together.

        The two matrices must have the same dimensions, otherwise a :class:`ValueError` is raised.
        Addition is done by going element by element, thus for :math:`C = A + B` we would have
        :math:`C_{i,j} = A_{i,j} + B_{i,j}` for all valid indices :math:`(i, j)`.

        Args:
            | other (:class:`Matrix`): The matrix to add.

        Returns:
            :class:`Matrix`: The sum of the matrices.
        """
        if self.rows != other.rows or self.cols != other.cols:
            raise ValueError(
                f"Cannot add matrices of different sizes. {self.rows}x{self.cols} + {other.rows}x{other.cols}"
            )

        entries = [x + y for (x, y) in zip(self.entries, other.entries)]
        return Matrix(self.rows, self.cols, entries)

    def __mul__(self, g: T | Matrix[T]) -> Matrix[T]:
        r"""Multiply a matrix.

        Two versions of this algorithm exist. One is multiplication by a scalar, the other is multiplication
        by a matrix. When multiplying by a scalar g, the multiplication is applied to each entry in the matrix
        such that if :math:`C = A * g` then :math:`C_{i,j} = A_{i,j} * g`. For multiplication by a matrix, the
        standard matrix multiplication algorithm is applied. If :math:`C = A * B` then :math:`C_{i,j}` is
        calculated as the dot product of the i-th row of A and the j-th column of B.

        Args:
            | g (:type:`T` | :class:`Matrix`): The scalar or matrix to multiply by.

        Returns:
            :class:`Matrix`: The result of the multiplication.
        """
        if isinstance(g, Matrix):
            # matrix multiplication
            if self.cols != g.rows:
                raise ValueError(
                    f"Matrix multiplication requires left hand matrix cols={self.cols} equal right hand matrix rows={g.rows}"
                )

            entries: list[T] = []
            for i in range(self.rows):
                for j in range(g.cols):
                    entry = self[(i, 0)] * g[(0, j)]
                    for k in range(1, g.rows):
                        entry += self[(i, k)] * g[(k, j)]
                    entries.append(entry)

            return Matrix(self.rows, g.cols, entries)
        else:
            # scalar multiplication
            entries = [u * g for u in self.entries]
            return Matrix(self.rows, self.cols, entries)

    def map(self, f: Callable[[T], T]) -> Matrix[T]:
        return Matrix(self.rows, self.cols, [f(x) for x in self.entries])

    def transpose(self) -> Matrix[T]:
        t = Matrix(self.cols, self.rows, self.entries[:])

        for i in range(self.rows):
            for j in range(self.cols):
                t[(j, i)] = self[(i, j)]

        return t

    def get_singleton_element(self) -> T:
        if len(self.entries) == 1:
            return self.entries[0]

        raise ValueError(
            f"Can only get singleton elements from 1x1 matrix (got {self.rows}x{self.cols})."
        )



from mlkem.auxiliary.crypto import XOF
from mlkem.auxiliary.general import bytes_to_bits
from mlkem.math.constants import n, q
from mlkem.math.field import Zm
from mlkem.math.polynomial_ring import PolynomialRing, RingRepresentation


def sample_ntt(b: bytes) -> PolynomialRing:
    if len(b) != 34:
        raise ValueError(
            f"Input must be 34 bytes (32-byte seed and two indices). Got {len(b)}."
        )

    a = PolynomialRing(representation=RingRepresentation.NTT)
    xof = XOF()
    xof.absorb(b)

    j = 0
    while j < n:
        c = xof.squeeze(3)
        d1 = c[0] + n * (c[1] % 16)
        d2 = c[1] // 16 + 16 * c[2]

        if d1 < q:
            a[j] = Zm(d1, q)
            j += 1

        if d2 < q and j < n:
            a[j] = Zm(d2, q)
            j += 1

    return a


def sample_poly_cbd(eta: int, b: bytes) -> PolynomialRing:
    if len(b) != 64 * eta:
        raise ValueError(f"Input must be {64 * eta} bytes (got {len(b)}).")

    f = PolynomialRing(representation=RingRepresentation.STANDARD)
    bits = bytes_to_bits([x for x in b])

    for i in range(n):
        x = sum([bits[2 * i * eta + j] for j in range(eta)])
        y = sum([bits[2 * i * eta + eta + j] for j in range(eta)])
        f[i] = Zm(x - y, q)

    return f



from dataclasses import dataclass


@dataclass
class ParameterSet:
    k: int
    eta1: int
    eta2: int
    du: int
    dv: int


# bit strength / security parameter = 128 bits
ML_KEM_512 = ParameterSet(k=2, eta1=3, eta2=2, du=10, dv=4)
# bit strength / security parameter = 192 bits
ML_KEM_768 = ParameterSet(k=3, eta1=2, eta2=2, du=10, dv=4)
# bit strength / security parameter = 256 bits
ML_KEM_1024 = ParameterSet(k=4, eta1=2, eta2=2, du=11, dv=5)



from hashlib import shake_128

from mlkem.auxiliary.crypto import g, prf
from mlkem.fastmath import (  # type: ignore
    add_matrix,
    add_poly,
    byte_decode_matrix,
    byte_decode_poly,
    byte_encode_matrix,
    byte_encode_poly,
    compress_matrix,
    compress_poly,
    decompress_matrix,
    decompress_poly,
    map_ntt_inv_matrix,
    map_ntt_matrix,
    mul_matrix,
    ntt_inv,
    sample_ntt,
    sample_poly_cbd,
    sub_poly,
)
from mlkem.parameter_set import ParameterSet


class Fast_K_PKE:
    """A public key encryption (PKE) scheme based on the module learning with errors (MLWE) problem."""

    def __init__(self, parameters: ParameterSet):
        self.parameters = parameters

    def key_gen(self, d: bytes) -> tuple[bytes, bytes]:
        r"""Creates a keypair used for encapsulation and decapsulation.

        The decryption (private) key is a vector :math:`s` of length k (where k is defined by the ML-KEM parameter set)
        with elements in :math:`R_q`. The encryption (public) key is a collection of "noisy" linear equations
        :math:`(A, As + e)` in the secret variable :math:`s`. The rows of matrix A, which is generated pseudorandomly,
        for the equation coeffients.

        Args:
            | d (:type:`bytes)`): The random seed used to derive the keypair. This should come from a random source suitable for cryptographic applications.

        Returns:
            :type:`tuple[bytes, bytes]`: The keypair with the encryption key first and the decryption key second.
        """
        k = self.parameters.k
        rho, sigma = g(d + bytes([k]))
        N = 0
        # generate matrix A in (Z^n_q)^{k*k}
        a_ = self._generate_a(rho)

        # generate vector s in (Z^n_q)^{k}
        s = self._sample_column_vector(self.parameters.eta1, sigma, N)
        N += k

        # generate vector e in (Z^n_q)^{k}
        e = self._sample_column_vector(self.parameters.eta1, sigma, N)
        N += k

        s_ = map_ntt_matrix(s)
        e_ = map_ntt_matrix(e)
        a_s_ = mul_matrix(a_, s_, k, k, k, 1)
        t_ = add_matrix(a_s_, e_)

        ek = byte_encode_matrix(t_, 12) + rho
        dk = byte_encode_matrix(s_, 12)
        return ek, dk

    def encrypt(self, ek: bytes, m: bytes, r: bytes) -> bytes:
        r"""Takes an encryption key ek, a 32 byte plaintext message m, and randomness r as input
        and produces a ciphertext c.

        The algorithm starts by deriving matrix A and vector t from the encryption key. It then
        generates a vector :math:`y \in R^k_q` and noise terms :math:`e1 \in R^k_q` and
        :math:`e2 \in R_q`. It then encodes the 256 bit plaintext as a polynomial with degree 255,
        where each bit of the plaintext is a coefficient of the polynomial. Then a new noisy equation
        is computed - :math:`(A^Ty + e_1, t^Ty, + e_2)`. An appropriate encoding of the message polynomial
        is then added to the latter term. Finally, the pair (u, v) is compressed and serialized into a byte
        array.

        Args:
            | ek (:type:`bytes)`: The encryption key.
            | m (:type:`bytes)`): The plaintext message.
            | r (:type:`bytes)`): The randomness.

        Returns:
            :type:`bytes)`: The ciphertext c = c1 + c2. c1 encodes u and c2 encodes v.

        """
        k = self.parameters.k
        du = self.parameters.du
        dv = self.parameters.dv
        N = 0

        # run byte_decode k times to decode t_ and extract 32 byte seed from ek
        t_ = byte_decode_matrix(ek[: 384 * k], 12, k)
        rho = ek[384 * k : 384 * k + 32]

        # regenerate matrix A that was sampled in key_gen
        a_ = self._generate_a(rho)
        # generate column vector y with entries sampled from CBD
        y = self._sample_column_vector(self.parameters.eta1, r, N)
        N += k
        # generate column vector e1 with entries sampled from CBD
        e1 = self._sample_column_vector(self.parameters.eta2, r, N)
        N += k
        e2 = sample_poly_cbd(
            prf(self.parameters.eta2, r, bytes([N])), self.parameters.eta2
        )

        y_ = map_ntt_matrix(y)
        a_y_ = mul_matrix(self._transpose(a_, k, k), y_, k, k, k, 1)
        u = add_matrix(map_ntt_inv_matrix(a_y_), e1)

        # encode plaintext m into polynomial v
        mu = decompress_poly(byte_decode_poly(m, 1), 1)
        t_y_ = mul_matrix(self._transpose(t_, k, 1), y_, 1, k, k, 1)
        ty = ntt_inv(t_y_[0])
        v = add_poly(add_poly(ty, e2), mu)

        # compress and encode c1 and c2
        c1 = byte_encode_matrix(compress_matrix(u, du), du)
        c2 = byte_encode_poly(compress_poly(v, dv), dv)

        return c1 + c2

    def decrypt(self, dk: bytes, c: bytes) -> bytes:
        du = self.parameters.du
        dv = self.parameters.dv
        k = self.parameters.k

        c1 = c[: 32 * du * k]
        c2 = c[32 * du * k : 32 * (du * k + dv)]

        # decode u, v and s
        u_prime = decompress_matrix(byte_decode_matrix(c1, du, k), du)
        v_prime = decompress_poly(byte_decode_poly(c2, dv), dv)
        s_ = byte_decode_matrix(dk, 12, k)

        # decode plaintext m from polynomial v
        s_u_ = mul_matrix(
            self._transpose(s_, k, 1), map_ntt_matrix(u_prime), 1, k, k, 1
        )
        w = sub_poly(v_prime, ntt_inv(s_u_[0]))
        m = byte_encode_poly(compress_poly(w, 1), 1)
        return m

    def _generate_a(self, rho: bytes) -> list[list[int]]:
        k = self.parameters.k
        result: list[list[int]] = []

        for i in range(k):
            for j in range(k):
                xof = shake_128()
                xof.update(rho + bytes([j, i]))

                # why 840? - # https://cryptojedi.org/papers/terminate-20230516.pdf
                element = sample_ntt(xof.digest(840))
                result.append(element)

        return result

    def _sample_column_vector(self, eta: int, r: bytes, N: int) -> list[list[int]]:
        """Generate a column vector in :math:`(Z^n_q)^{k}"""
        v: list[list[int]] = []

        for _ in range(self.parameters.k):
            seed = prf(eta, r, bytes([N]))
            v.append(sample_poly_cbd(seed, eta))
            N += 1

        return v

    def _transpose(self, m: list[list[int]], rows: int, cols: int) -> list[list[int]]:
        t: list[list[int]] = [[] for _ in range(rows * cols)]
        for i in range(rows):
            for j in range(cols):
                t[j * rows + i] = m[i * cols + j]
        return t



from binascii import hexlify
from functools import reduce
from logging import getLogger

from mlkem.auxiliary.crypto import g, prf
from mlkem.auxiliary.general import (
    byte_decode,
    byte_encode,
    compress,
    decompress,
)
from mlkem.auxiliary.ntt import ntt, ntt_inv
from mlkem.auxiliary.sampling import sample_ntt, sample_poly_cbd
from mlkem.math.field import Zm
from mlkem.math.matrix import Matrix
from mlkem.math.polynomial_ring import PolynomialRing, RingRepresentation
from mlkem.parameter_set import ParameterSet

LOG = getLogger(__name__)


class K_PKE:
    """A public key encryption (PKE) scheme based on the module learning with errors (MLWE) problem."""

    def __init__(self, parameters: ParameterSet):
        self.parameters = parameters

    def key_gen(self, d: bytes) -> tuple[bytes, bytes]:
        r"""Creates a keypair used for encapsulation and decapsulation.

        The decryption (private) key is a vector :math:`s` of length k (where k is defined by the ML-KEM parameter set)
        with elements in :math:`R_q`. The encryption (public) key is a collection of "noisy" linear equations
        :math:`(A, As + e)` in the secret variable :math:`s`. The rows of matrix A, which is generated pseudorandomly,
        for the equation coeffients.

        Args:
            | d (:type:`bytes)`): The random seed used to derive the keypair. This should come from a random source suitable for cryptographic applications.

        Returns:
            :type:`tuple[bytes, bytes]`: The keypair with the encryption key first and the decryption key second.
        """
        k = self.parameters.k
        eta1 = self.parameters.eta1
        rho, sigma = g(d + bytes([k]))
        N = 0
        LOG.debug(f"rho: {hexlify(rho).decode()}")
        LOG.debug(f"sigma: {hexlify(sigma).decode()}")

        # generate matrix A in (Z^n_q)^{k*k}
        a_ = self._generate_a(rho)
        LOG.debug(f"aHat: {a_}")

        # generate vector s in (Z^n_q)^{k}
        s, N = self._sample_column_vector(eta1, sigma, N)
        LOG.debug(f"s: {s}")

        # generate vector e in (Z^n_q)^{k}
        e, N = self._sample_column_vector(eta1, sigma, N)
        LOG.debug(f"e: {e}")

        s_ = s.map(ntt)
        LOG.debug(f"sHat: {s_}")
        e_ = e.map(ntt)
        LOG.debug(f"eHat: {e_}")
        a_s_ = a_ * s_
        LOG.debug(f"aHat * sHat: {a_s_}")
        t_ = a_s_ + e_
        LOG.debug(f"tHat = aHat * sHat + eHat:: {t_}")

        # use reduce to map byte encoding over the vectors
        # 1. start with an empty byte sequence
        # 2. grab the next entry in the vector
        # 3. apply byte_encode to the coefficients of the entry (the entry is a polynomial ring element)
        # 4. append the bytes from encoding to the end of the byte sequence
        ek = (
            reduce(
                lambda x, y: x + byte_encode(12, y.coefficients),
                t_.entries,
                b"",
            )
            + rho
        )
        dk = reduce(
            lambda x, y: x + byte_encode(12, y.coefficients),
            s_.entries,
            b"",
        )
        LOG.debug(f"ek: {hexlify(ek).decode()}")
        LOG.debug(f"dk: {hexlify(dk).decode()}")

        return ek, dk

    def encrypt(self, ek: bytes, m: bytes, r: bytes) -> bytes:
        r"""Takes an encryption key ek, a 32 byte plaintext message m, and randomness r as input
        and produces a ciphertext c.

        The algorithm starts by deriving matrix A and vector t from the encryption key. It then
        generates a vector :math:`y \in R^k_q` and noise terms :math:`e1 \in R^k_q` and
        :math:`e2 \in R_q`. It then encodes the 256 bit plaintext as a polynomial with degree 255,
        where each bit of the plaintext is a coefficient of the polynomial. Then a new noisy equation
        is computed - :math:`(A^Ty + e_1, t^Ty, + e_2)`. An appropriate encoding of the message polynomial
        is then added to the latter term. Finally, the pair (u, v) is compressed and serialized into a byte
        array.

        Args:
            | ek (:type:`bytes)`: The encryption key.
            | m (:type:`bytes)`): The plaintext message.
            | r (:type:`bytes)`): The randomness.

        Returns:
            :type:`bytes)`: The ciphertext c = c1 + c2. c1 encodes u and c2 encodes v.

        """
        k = self.parameters.k
        du = self.parameters.du
        dv = self.parameters.dv
        N = 0

        # run byte_decode k times to decode t_ and extract 32 byte seed from ek
        t_ = self._bytes_to_column_vector(ek[: 384 * k], RingRepresentation.NTT, 12)
        rho = ek[384 * k : 384 * k + 32]

        # regenerate matrix A that was sampled in key_gen
        a_ = self._generate_a(rho)
        # generate column vector y with entries sampled from CBD
        y, N = self._sample_column_vector(self.parameters.eta1, r, N)
        # generate column vector e1 with entries sampled from CBD
        e1, N = self._sample_column_vector(self.parameters.eta2, r, N)
        e2 = sample_poly_cbd(
            self.parameters.eta2, prf(self.parameters.eta2, r, bytes([N]))
        )

        y_ = y.map(ntt)
        u = (a_.transpose() * y_).map(ntt_inv) + e1

        # encode plaintext m into polynomial v
        mu = PolynomialRing(
            [decompress(1, x) for x in byte_decode(1, m)], RingRepresentation.STANDARD
        )
        v = ntt_inv((t_.transpose() * y_).get_singleton_element()) + e2 + mu

        # compress and encode c1 and c2
        compressed_u: list[list[Zm]] = reduce(
            lambda x, y: x + [[compress(du, z) for z in y.coefficients]],
            u.entries,
            [],
        )
        c1 = reduce(
            lambda x, y: x + byte_encode(du, y),
            compressed_u,
            b"",
        )
        c2 = byte_encode(dv, [compress(dv, x) for x in v.coefficients])

        return c1 + c2

    def decrypt(self, dk: bytes, c: bytes) -> bytes:
        du = self.parameters.du
        dv = self.parameters.dv
        k = self.parameters.k

        c1 = c[: 32 * du * k]
        c2 = c[32 * du * k : 32 * (du * k + dv)]

        # decode u, v and s
        u_prime = self._bytes_to_column_vector(
            c1, RingRepresentation.STANDARD, du, compressed=True
        )
        v_prime = PolynomialRing(
            [decompress(dv, x) for x in byte_decode(dv, c2)],
            RingRepresentation.STANDARD,
        )
        s_ = self._bytes_to_column_vector(dk, RingRepresentation.NTT, 12)

        # decode plaintext m from polynomial v
        w = v_prime - ntt_inv(
            (s_.transpose() * u_prime.map(ntt)).get_singleton_element()
        )
        m = byte_encode(1, [compress(1, x) for x in w.coefficients])
        return m

    def _generate_a(self, rho: bytes) -> Matrix[PolynomialRing]:
        k = self.parameters.k
        a_ = Matrix(
            rows=k,
            cols=k,
            constructor=lambda: PolynomialRing(representation=RingRepresentation.NTT),
        )
        for i in range(k):
            for j in range(k):
                a_[(i, j)] = sample_ntt(rho + bytes([j, i]))

        return a_

    def _sample_column_vector(
        self, eta: int, r: bytes, N: int
    ) -> tuple[Matrix[PolynomialRing], int]:
        """Generate a column vector in :math:`(Z^n_q)^{k}"""
        v: Matrix[PolynomialRing] = Matrix(
            rows=self.parameters.k,
            cols=1,
            constructor=lambda: PolynomialRing(
                representation=RingRepresentation.STANDARD
            ),
        )
        for i in range(self.parameters.k):
            seed = prf(eta, r, bytes([N]))
            # vectors are columnar, so column index is always 0
            v[(i, 0)] = sample_poly_cbd(eta, seed)
            N += 1

        return v, N

    def _bytes_to_column_vector(
        self,
        b: bytes,
        representation: RingRepresentation,
        d: int,
        compressed: bool = False,
    ) -> Matrix[PolynomialRing]:
        k = self.parameters.k
        coefficient_size = 32 * d

        return Matrix(
            rows=k,
            cols=1,
            entries=[
                PolynomialRing(
                    [
                        decompress(d, x)
                        for x in byte_decode(d, b[i : i + coefficient_size])
                    ]
                    if compressed
                    else byte_decode(d, b[i : i + coefficient_size]),
                    representation,
                )
                for i in range(0, coefficient_size * k, coefficient_size)
            ],
        )



from setuptools import Extension, setup  # type: ignore

fastmath = Extension(
    "mlkem.fastmath",
    sources=["mlkem/math/fastmathmodule.c"],
    extra_compile_args=["-std=c99"],
)

setup(
    ext_modules=[fastmath],
)



from unittest import TestCase

from mlkem.ml_kem import ML_KEM
from mlkem.parameter_set import ML_KEM_512


class TestML_KEM(TestCase):
    def test_full(self) -> None:
        ml_kem = ML_KEM(ML_KEM_512)

        ek, dk = ml_kem.key_gen()
        k, c = ml_kem.encaps(ek)
        k_ = ml_kem.decaps(dk, c)

        self.assertEqual(k, k_)



from __future__ import annotations
from enum import StrEnum

from mlkem.math.constants import n, q
from mlkem.math.field import Zm


class RingRepresentation(StrEnum):
    STANDARD = "Standard Polynomial Ring (Rq)"
    NTT = "NTT Representation (Tq)"


class PolynomialRing:
    r"""Represents elements of the ring :math:`\mathbb{Z}^n_q`."""

    coefficients: list[Zm]
    representation: RingRepresentation

    def __init__(
        self,
        coefficients: list[Zm] | None = None,
        representation: RingRepresentation = RingRepresentation.STANDARD,
    ):
        if coefficients is None:
            coefficients = [Zm(0, q) for _ in range(n)]

        if len(coefficients) != n:
            raise ValueError(f"coefficients must have length {n}")

        self.coefficients = coefficients
        self.representation = representation

    def __eq__(self, other: object) -> bool:
        """Compare two polynomial rings for equality.

        They must have the following conditions met to be equal:
          * The same representation
          * The same coefficients
          * The same order (length of coefficients)

        Args:
            | other (:class:`object`): The object to compare to (must be a :class:`PolynomialRing` instance).

        Returns:
            :type:`bool`: Whether the two objects are equal.
        """
        if not isinstance(other, PolynomialRing):
            return NotImplemented

        if self.representation != other.representation:
            return False

        if len(self.coefficients) != len(other.coefficients):
            return False

        return all(
            [fi == gi for (fi, gi) in zip(self.coefficients, other.coefficients)]
        )

    def __repr__(self) -> str:
        return "[ " + ", ".join([repr(x) for x in self.coefficients]) + " ]"

    def __getitem__(self, index: int) -> Zm:
        """Get the coefficient at :code:`index`."""
        if index >= n:
            raise IndexError(
                f"Index for Rq coefficient must be less than {n}. Got {index}."
            )

        return self.coefficients[index]

    def __setitem__(self, index: int, value: Zm) -> None:
        """Set the coefficient at :code:`index` to :code:`value`."""
        if index >= n:
            raise IndexError(
                f"Index for Rq coefficient must be less than {n}. Got {index}."
            )

        self.coefficients[index] = value

    def __add__(self, g: PolynomialRing) -> PolynomialRing:
        r"""Add two elements f and g where :math:`f, g \in \mathbb{Z}^n_m`.

        The resulting element will also be in :math:`\mathbb{Z}^n_m`.

        Args:
            | g (:class:`PolynomialRing`): The element to add to f (represent by self).

        Returns:
            :class:`PolynomialRing`: An element equal to :math:`f + g`.
        """
        if self.representation != g.representation:
            raise ValueError(
                "PolynomialRing representation must be the same in order to perform arithmetic."
            )

        coefficients = [fi + gi for (fi, gi) in zip(self.coefficients, g.coefficients)]
        return PolynomialRing(coefficients, representation=self.representation)

    def __sub__(self, g: PolynomialRing) -> PolynomialRing:
        r"""Subtract two elements f and g where :math:`f, g \in \mathbb{Z}^n_m`.

        The resulting element will also be in :math:`\mathbb{Z}^n_m`.

        Args:
            | g (:class:`PolynomialRing`): The element to subtract from f (represent by self).

        Returns:
            :class:`PolynomialRing`: An element equal to :math:`f - g`.
        """
        if self.representation != g.representation:
            raise ValueError(
                "PolynomialRing representation must be the same in order to perform arithmetic."
            )

        coefficients = [fi - gi for (fi, gi) in zip(self.coefficients, g.coefficients)]
        return PolynomialRing(coefficients, representation=self.representation)

    def __mul__(self, a: Zm | PolynomialRing) -> PolynomialRing:
        r"""Multiply by an element in :math:`\mathbb{Z}^m` or :math:`\mathbb{Z}^n_m`.

        If :math:`a \in \mathbb{Z}_m` and :math:`f \in \mathbb{Z}^n_m` the i-th coefficient of the polynomial
        :math:`a \cdot f \in \mathbb{Z}^n_m` is equal to :math:`a \cdot f_i \pmod{m}`.

        If :math:`a \in \mathbb{Z}^n_m` and :math:`f \in \mathbb{Z}^n_m` then both polynomials must be in NTT
        representation. NTT multiplication is then used to multiply the two together.

        Args:
            | a (:class:`Zm` | :class:`PolynomialRing`): The value to multiply by.

        Returns:
            :class:`PolynomialRing`: The element in :math:`\mathbb{Z}^n_m` multiplied by `a`.
        """
        # scalar multiplication
        if isinstance(a, Zm):
            coefficients = [a * fi for fi in self.coefficients]
            return PolynomialRing(coefficients, representation=self.representation)

        # polynomial multiplication
        elif isinstance(a, PolynomialRing):
            from mlkem.auxiliary.ntt import multiply_ntt

            if (
                self.representation != RingRepresentation.NTT
                or a.representation != RingRepresentation.NTT
            ):
                raise ValueError(
                    "Multiplying PolynomialRings is only possible if both are in NTT representation."
                )

            return multiply_ntt(self, a)

        else:
            raise NotImplementedError(
                f"Cannot multiply PolynomialRing by type {type(a)}."
            )

    def __rmul__(self, a: Zm | PolynomialRing) -> PolynomialRing:
        r"""Equivalent to :code:`self.__mul__(a)`."""
        return self.__mul__(a)
